from typing import Optional, Callable, List, Union
from abc import ABC
import os

from .config import GenerationConfig, SessionConfig, ContainerType, QuantizationType
from .results import GenerationResult



#Theoretically this is incorrect as the 'model' doesnt actually exist, but it is a good enough approximation for now. 
class Model(ABC):
    """
    Wrapper around a llm model.
    """
    config:SessionConfig
    
    @property
    def path(self)->str: ...
    
    @property
    def verbose(self)->bool: ...

    @property
    def lora_paths(self)->Optional[List[str]]: ...

    def  __init__(self,
                  path:Union[str,os.PathLike],
                  session_config:SessionConfig=SessionConfig(),
                  lora_paths:Optional[List[Union[str,os.PathLike]]]=None,
                  verbose:bool=False) -> None: ...
    
    def generate(self,prompt:str,
                 generation_config:Optional[GenerationConfig]=None,
                 callback:Callable[[str],Optional[bool]]=None) -> GenerationResult: 
        """
        Generates text from a prompt.
        """ 
        ...
    
    def tokenize(self,text:str) -> List[int]:
        """
        Tokenizes a string into a list of tokens.
        """
        ...
    
    def decode(self,tokens:List[int]) -> str:
        """
        Decodes a list of tokens into a string.
        """
        ...

    @staticmethod
    def quantize(source:str,destination:str,quantization:QuantizationType=QuantizationType.Q4_0,container:ContainerType=ContainerType.GGJT)->None:
        """
        Quantizes the model.
        """
        ...


import numpy as np
from ._base import BaseAdapter
from typing import Union,Tuple,Optional,BinaryIO 
from transformers import AutoTokenizer,AutoModelForCausalLM,AutoConfig,GPT2LMHeadModel
import os
import torch
import struct
import numpy as np
from ...auto import KnownModels
import re

#based on https://github.com/ggerganov/ggml/blob/master/examples/gpt-2/convert-h5-to-ggml.py
class Gpt2Converter(BaseAdapter):
    model_type:KnownModels=KnownModels.Gpt2

    def load(self,pretrained_model_name_or_path:Union[str,os.PathLike],pretrained_tokenizer_name_or_path:Optional[Union[str,os.PathLike]]=None)->Tuple[AutoTokenizer,AutoModelForCausalLM]:
        config = AutoConfig.from_pretrained(pretrained_model_name_or_path)
        
        model = GPT2LMHeadModel.from_pretrained(
            pretrained_model_name_or_path,
            torch_dtype=torch.float16,
            low_cpu_mem_usage=True,
            )
        tokenizer_name = pretrained_tokenizer_name_or_path if pretrained_tokenizer_name_or_path else pretrained_model_name_or_path 
        tokenizer = AutoTokenizer.from_pretrained(tokenizer_name)
        return config,tokenizer,model
    
    def _write_hyperparameters(self,out_file:BinaryIO):
        hyperparameters = self.config.to_dict()
        out_file.write(struct.pack("i", hyperparameters["vocab_size"]))
        out_file.write(struct.pack("i", hyperparameters["n_positions"]))
        out_file.write(struct.pack("i", hyperparameters["n_embd"]))
        out_file.write(struct.pack("i", hyperparameters["n_head"]))
        out_file.write(struct.pack("i", hyperparameters["n_layer"]))


    def _filter_weights(self, name: str, weight: torch.Tensor) -> bool:
        return name.endswith("attn.masked_bias") or name.endswith(".attn.bias") 
    
    def _rename_weights(self, name: str) -> str:
        new_name=name
        # rename headers to keep compatibility
        # cerebras renames, see https://github.com/ggerganov/ggml/blob/master/examples/gpt-2/convert-cerebras-to-ggml.py
        if name == "transformer.ln_f.weight":
            new_name = "model/ln_f/g"
        elif name == "transformer.ln_f.bias":
            new_name = "model/ln_f/b"
        elif name == "transformer.wte.weight":
            new_name = "model/wte"
        elif name == "transformer.wpe.weight":
            new_name = "model/wpe"
        elif name == "lm_head.weight":
            new_name = "model/lm_head"
        elif re.match(r"transformer.h\.\d+\.ln_1\.weight", name):
            i = re.findall("\d+", name)[0]
            new_name = f"model/h{i}/ln_1/g"
        elif re.match(r"transformer.h\.\d+\.ln_1\.bias", name):
            i = re.findall("\d+", name)[0]
            new_name = f"model/h{i}/ln_1/b"
        elif re.match(r"transformer.h\.\d+\.attn\.c_attn\.weight", name):
            i = re.findall("\d+", name)[0]
            new_name = f"model/h{i}/attn/c_attn/w"
        elif re.match(r"transformer.h\.\d+\.attn\.c_attn\.bias", name):
            i = re.findall("\d+", name)[0]
            new_name = f"model/h{i}/attn/c_attn/b"
        elif re.match(r"transformer.h\.\d+\.attn\.c_proj\.weight", name):
            i = re.findall("\d+", name)[0]
            new_name = f"model/h{i}/attn/c_proj/w"
        elif re.match(r"transformer.h.\d+.attn.c_proj.bias", name):
            i = re.findall("\d+", name)[0]
            new_name = f"model/h{i}/attn/c_proj/b"
        elif re.match(r"transformer.h.\d+.ln_2.weight", name):
            i = re.findall("\d+", name)[0]
            new_name = f"model/h{i}/ln_2/g"
        elif re.match(r"transformer.h.\d+.ln_2.bias", name):
            i = re.findall("\d+", name)[0]
            new_name = f"model/h{i}/ln_2/b"
        elif re.match(r"transformer.h.\d+.mlp.c_fc.weight", name):
            i = re.findall("\d+", name)[0]
            new_name = f"model/h{i}/mlp/c_fc/w"
        elif re.match(r"transformer.h.\d+.mlp.c_fc.bias", name):
            i = re.findall("\d+", name)[0]
            new_name = f"model/h{i}/mlp/c_fc/b"
        elif re.match(r"transformer.h.\d+.mlp.c_proj.weight", name):
            i = re.findall("\d+", name)[0]
            new_name = f"model/h{i}/mlp/c_proj/w"
        elif re.match(r"transformer.h.\d+.mlp.c_proj.bias", name):
            i = re.findall("\d+", name)[0]
            new_name = f"model/h{i}/mlp/c_proj/b"

        return new_name
    
    def _write_vocabulary(self, out_file: BinaryIO):
        # write the vocabulary size
        out_file.write(struct.pack("i", self.config.vocab_size))
        return super()._write_vocabulary(out_file)

    def _transform_weights(self, name: str, weight: torch.Tensor) -> torch.Tensor:
        # for efficiency - transpose these matrices:
        if name.endswith("/mlp/c_proj/w") or name.endswith("/mlp/c_fc/w") or name.endswith("/attn/c_proj/w") or name.endswith("/attn/c_attn/w"):
            #see this issue: https://github.com/pytorch/pytorch/issues/50275
            return weight.transpose(0,1)
        return weight
    
    def _filter_f16_weights(self, name: str, data: np.ndarray) -> bool:
        n_dims = len(data.shape)
        return (name == "model/wte" or name == "model/lm_head" or name[-2:] == "/g" or name[-2:] == "/w") and n_dims == 2

    

        


from .config import QuantizationType,ContainerType,SessionConfig
import os
import pathlib
from .models import Mpt,GptNeoX,GptJ,Gpt2,Bloom,Llama
from .base_model import Model
import logging
from typing import Optional, List, Union,Type
import os
from enum import Enum, auto
from dataclasses import dataclass
import json
from blake3 import blake3

class KnownModels(Enum):
    GptNeoX = auto()
    Mpt = auto()
    GptJ = auto()
    Gpt2 = auto()
    Bloom = auto()
    Llama = auto()


_QUANTIZATION_TYPE_MAP = {
    "Q4_0": QuantizationType.Q4_0,
    "Q4_1": QuantizationType.Q4_1,
    "F16": QuantizationType.F16
}

_CONTAINER_TYPE_MAP = {
    "GGML": ContainerType.GGML,
    "GGJT": ContainerType.GGJT
}

_KNOWN_MODELS_MAP = {
    KnownModels.GptNeoX: GptNeoX,
    KnownModels.Mpt: Mpt,
    KnownModels.GptJ: GptJ,
    KnownModels.Gpt2: Gpt2,
    KnownModels.Bloom: Bloom,
    KnownModels.Llama: Llama
}

@dataclass()
class ModelMetadata():
    """
    A dataclass to store metadata about a model.
    """
    model: KnownModels
    quantization: QuantizationType
    container: ContainerType
    converter:str="llm-rs"
    hash:Optional[str]=None

    def add_hash(self,model_path:Union[str,os.PathLike]):
        h  = blake3(max_threads=blake3.AUTO)
        b  = bytearray(128_000_000)
        mv = memoryview(b)
        with open(model_path, 'rb', buffering=0) as f:
            for n in iter(lambda : f.readinto(mv), 0):
                h.update(mv[:n])
        self.hash=h.hexdigest()

    def serialize(self):
        return {
            "model": self.model.name,
            "quantization": repr(self.quantization).split(".")[-1],
            "container": repr(self.container).split(".")[-1],
            "converter": self.converter,
            "hash": self.hash
        }
    
    @staticmethod
    def deserialize(metadata_dict):
        return ModelMetadata(
            model = KnownModels[metadata_dict["model"]],
            quantization = _QUANTIZATION_TYPE_MAP[metadata_dict["quantization"]],
            container = _CONTAINER_TYPE_MAP[metadata_dict["container"]],
            converter = metadata_dict["converter"],
            hash = metadata_dict["hash"]
        )




class AutoModel():
    """
    Utility to load models, without having to specify the model type.
    """
    @staticmethod
    def load_metadata(model_file:Union[str,os.PathLike])->ModelMetadata:
        path = pathlib.Path(model_file)
        if not path.is_file():
            raise ValueError(f"Model file '{model_file}' is not a file!")
        if not path.exists():
            raise ValueError(f"Model file '{model_file}' does not exist!")

        metadata_file = path.with_suffix(".meta")
        if not metadata_file.exists():
            raise ValueError(f"Model file '{model_file}' does not have a metadata file '{metadata_file}'!")
        
        metadata = ModelMetadata.deserialize(json.loads(metadata_file.read_text()))
        return metadata
    
    @staticmethod
    def infer_model_type(model_file:Union[str,os.PathLike])->Type[Model]:
        
        metadata = AutoModel.load_metadata(model_file)
        if metadata.model in _KNOWN_MODELS_MAP:
            return _KNOWN_MODELS_MAP[metadata.model]
        else:
            raise ValueError(f"Unknown model type '{metadata.model}'")
            
        
    @staticmethod
    def load(path:Union[str,os.PathLike],
                  session_config:SessionConfig=SessionConfig(),
                  lora_paths:Optional[List[Union[str,os.PathLike]]]=None,
                  verbose:bool=False)->Model:
        
        model_type = AutoModel.infer_model_type(path)
        return model_type(path,session_config,lora_paths,verbose)
            

class AutoQuantizer():
    """
    Utility to quantize models, without having to specify the model type.
    """
    @staticmethod
    def quantize(model_file:Union[str,os.PathLike],target_path:Optional[Union[str,os.PathLike]]=None,quantization:QuantizationType=QuantizationType.Q4_0,container:ContainerType=ContainerType.GGJT)->Union[str,os.PathLike]:
        metadata=AutoModel.load_metadata(model_file)
        if metadata.quantization != QuantizationType.F16:
            raise ValueError(f"Model '{model_file}' is already quantized to '{metadata.quantization}'")

        model_type = AutoModel.infer_model_type(model_file)

        if target_path is None:
            target_path = os.path.dirname(model_file)
            
        def build_target_name()->str:
            output_path = pathlib.Path(target_path)
            if output_path.is_file():
                return output_path
            else:
                output_path.mkdir(parents=True,exist_ok=True)
                model_path = pathlib.Path(model_file)
                appendix = ""
                if  quantization == QuantizationType.Q4_0:
                    appendix += "-q4_0"
                elif quantization == QuantizationType.Q4_1:
                    appendix += "-q4_1"

                if container == ContainerType.GGJT:
                    appendix += "-ggjt"
            
                filename = model_path.stem.replace("-f16","") + appendix + model_path.suffix
                return str(output_path / filename)

        target_file = build_target_name()
        if pathlib.Path(target_file).exists():
            logging.warning(f"Target file '{target_file}' already exists, skipping quantization")
            return target_file
        
        logging.info(f"Quantizing model '{model_file}' to '{target_file}'")
        model_type.quantize(model_file,target_file,quantization,container)

        metadata_file = pathlib.Path(target_file).with_suffix(".meta")
        quantized_metadata = ModelMetadata(model=metadata.model,quantization=quantization,container=container)
        quantized_metadata.add_hash(target_file)
        logging.info(f"Writing metadata file '{metadata_file}'")
        metadata_file.write_text(json.dumps(quantized_metadata.serialize()))
        logging.info(f"Finished quantizing model '{model_file}' to '{target_file}'")
        return target_file





from .llm_rs import *

from .config import GenerationConfig, SessionConfig, Precision, ContainerType, QuantizationType
from .models import Llama, GptJ, Gpt2, Bloom, GptNeoX, Mpt
from .auto import AutoModel, ModelMetadata, KnownModels, AutoQuantizer

__doc__ = llm_rs.__doc__
if hasattr(llm_rs, "__all__"):
    __all__ = llm_rs.__all__


from .base_model import Model
    
class Llama(Model):
    """
    Wrapper around all Llama based models.
    """
    ...
    
    
class GptJ(Model):
    """
    Wrapper around all GPTJ based models.
    """
    ...
    
class Gpt2(Model):
    """
    Wrapper around all GPT2 based models.
    """
    ...
    
class Bloom(Model):
    """
    Wrapper around all Bloom based models.
    """
    ...
    
class GptNeoX(Model):
    """
    Wrapper around all GPT-NeoX based models.
    """
    ...

class Mpt(Model):
    """
    Wrapper around all Mpt based models.
    """
    ...


from transformers import AutoConfig
import os
from typing import Union
from .models import MPTConverter,GptNeoXConverter,GptJConverter,Gpt2Converter,BloomConverter,LlamaConverter
import logging 
import pathlib


_ARCHITECTURE_CONVERTER_MAP = {
    "MPTForCausalLM": MPTConverter,
    "GPTNeoXForCausalLM": GptNeoXConverter,
    "GPTJForCausalLM": GptJConverter,
    "GPT2LMHeadModel": Gpt2Converter,
    "BloomForCausalLM":BloomConverter,
    "LLaMAForCausalLM":LlamaConverter
}

class AutoConverter():

    @staticmethod
    def convert(pretrained_model_name_or_path:Union[str,os.PathLike],output_path:Union[str,os.PathLike])->str:
        config = AutoConfig.from_pretrained(pretrained_model_name_or_path,trust_remote_code=True)
        architecture = config.architectures[0]
        model_name = get_name_from_config(config)

        adapter=None
        if architecture not in _ARCHITECTURE_CONVERTER_MAP:
            raise ValueError(f"Unsupported architecture '{architecture}' for '{model_name}'")
        adapter = _ARCHITECTURE_CONVERTER_MAP[architecture]

        output_file = build_path(output_path,model_name)
        if os.path.exists(output_file):
            logging.warning(f"Skipping {model_name} via {adapter.__name__} because '{output_file}' already exists!")
            return output_file
        
        logging.info(f"Converting {model_name} via {adapter.__name__} to '{output_file}'")
        adapter(pretrained_model_name_or_path).convert(output_file)
        return output_file


def get_name_from_config(config:AutoConfig)->str:
    raw_name:str = ""
    if hasattr(config,"name_or_path"):
        raw_name = config.name_or_path
    elif hasattr(config,"model_type"):
        raw_name = config.model_type
    else:
        raw_name = config.architectures[0]
    return raw_name.split("/")[-1]


def build_path(output_path:Union[str,os.PathLike],model_name:str)->str:
    output_path = pathlib.Path(output_path)
    #User specified a file, so just use that
    if output_path.is_file():
        return output_path

    #User specified a directory => auto generate a file name
    output_path.mkdir(parents=True,exist_ok=True)
    return os.path.join(output_path,f"{model_name}-f16.bin")
    
    


from abc import ABC, abstractmethod
from typing import Union,Tuple,Any,Optional,BinaryIO
from transformers import AutoTokenizer,AutoModelForCausalLM,AutoConfig
import os
import struct
from enum import Enum
import logging
import numpy as np
from ...config import QuantizationType,ContainerType
from ...auto import ModelMetadata, KnownModels
import pathlib
import json
import torch

GGML_MAGIC = 0x67676D6C
GGMF_MAGIC = 0x67676D66
GGJT_MAGIC = 0x67676a74

class FileTypes(Enum):
    FP32 = 0
    FP16 = 1

class BaseAdapter(ABC):
    model_type:KnownModels=None
    file_magic:int=GGML_MAGIC
    version:int=1

    def  __init__(self,pretrained_model_name_or_path:Union[str,os.PathLike],pretrained_tokenizer_name_or_path:Optional[Union[str,os.PathLike]]=None) -> None:
        self.config,self.tokenizer,self.model= self.load(pretrained_model_name_or_path,pretrained_tokenizer_name_or_path)

    @abstractmethod
    def load(self,pretrained_model_name_or_path:Union[str,os.PathLike],pretrained_tokenizer_name_or_path:Optional[Union[str,os.PathLike]]=None)->Tuple[AutoConfig,AutoTokenizer,AutoModelForCausalLM]:
        ...

    @abstractmethod
    def _write_hyperparameters(self,out_file:BinaryIO):
        ...

    def _write_vocabulary(self,out_file:BinaryIO):
        # TODO: temporary hack to not deal with implementing the tokenizer
        logging.info(f"Processing vocabulary with size {self.config.vocab_size}")
        dot_token = self.tokenizer.encode(".")[0]
        for i in range(self.config.vocab_size):
            text = self.tokenizer.decode([dot_token, i]).encode("utf-8")
            # remove the first byte (it's always '.')
            text = text[1:]
            out_file.write(struct.pack("i", len(text)))
            out_file.write(text)


    def _filter_weights(self,name:str,weight:torch.Tensor)->bool:
        """Filter weights that should be skipped"""
        return False
    
    def _filter_weights_after_rename(self,name:str,weight:torch.Tensor)->bool:
        """Filter weights that should be skipped"""
        return False
    
    def _filter_f16_weights(self,name:str,data:np.ndarray)->bool:
        """Filter weights that should be stored as fp16"""
        n_dims = len(data.shape)
        return name.endswith(".weight") and n_dims == 2
    
    def _rename_weights(self,name:str)->str:
        """Rename weights that should be renamed"""
        return name
    
    def _transform_weights(self,name:str,weight:torch.Tensor)->torch.Tensor:
        """Transform weights that should be transformed"""
        return weight
    
    def _write_weights(self,out_file:BinaryIO):
        weights = self.model.state_dict()
        for name, weight in weights.items():
            

            if self._filter_weights(name,weight):
                logging.info(f"Skipping layer '{name}'")
                continue
            
            name = self._rename_weights(name)
            weight = self._transform_weights(name,weight)
            if self._filter_weights_after_rename(name,weight):
                logging.info(f"Skipping layer '{name}'")
                continue

            data = weight.squeeze().numpy()
            n_dims = len(data.shape);    
            type = FileTypes.FP32

            if self._filter_f16_weights(name,data):
                data = data.astype(np.float16)
                type = FileTypes.FP16
            else:
                data = data.astype(np.float32)

            encoded_name = name.encode("utf-8")
            out_file.write(struct.pack("iii", n_dims, len(encoded_name), type.value))

            if self.file_magic == GGJT_MAGIC or self.file_magic == GGMF_MAGIC:
                out_file.write(struct.pack("i" * len(data.shape), *data.shape[::-1]))
            else:
                for i in range(n_dims):
                    out_file.write(struct.pack("i", data.shape[n_dims - 1 - i]))
            out_file.write(encoded_name)


            if self.file_magic == GGJT_MAGIC:
                # pad to 32 bytes
                out_file.seek((out_file.tell() + 31) & -32)

            # data
            data.tofile(out_file)
            logging.info(f"Converted layer '{name}' with shape {data.shape}")


    def write_magic(self,out_file:BinaryIO):
        out_file.write(struct.pack("i", self.file_magic))
        if self.file_magic == GGJT_MAGIC or self.file_magic == GGMF_MAGIC:
            out_file.write(struct.pack("i", self.version))

    def write_file_type(self,out_file:BinaryIO,file_type:FileTypes=FileTypes.FP16):
        out_file.write(struct.pack("i", file_type.value))



    def convert(self,output_file:Union[str,os.PathLike])->None:

        with open(output_file, "wb") as out_file:
            # write magic
            self.write_magic(out_file)
            logging.info(f"Processing hyperparameters ...")
            self._write_hyperparameters(out_file)
            self.write_file_type(out_file)

            self._write_vocabulary(out_file)
            self._write_weights(out_file)
            logging.info(f"Done converting model to GGML format. Saved to '{output_file}'")

        #Create the *.meta file needed for automatic loading
        metadata_file = pathlib.Path(output_file).with_suffix(".meta")
        metadata = ModelMetadata(
            self.model_type,
            QuantizationType.F16,
            ContainerType.GGML
            )
        metadata.add_hash(output_file)
        metadata_file.write_text(json.dumps(metadata.serialize(),indent=4))
        
        logging.info(f"Created metadata file at '{output_file}'")
        return output_file

